﻿using Newtonsoft.Json.Linq;
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Formatting;
using System.Net.Http.Headers;
using System.Text;
using System.Threading.Tasks;

namespace RavenDbMapReduceSort
{
    public interface IServiceChannel
    {
        TResponse Get<TResponse>(string url, HeadersCollection headers = null);
        TResponse Post<TResponse>(JToken obj, string url, HeadersCollection headers);
        TResponse Put<TResponse>(JToken obj, string url, HeadersCollection headers);
    }

    public class ServiceChannel : IServiceChannel
    {
        public readonly string SERVICE_URL;
        public ServiceChannel(string serviceUrl)
        {
            SERVICE_URL = serviceUrl;
        }

        public TResponse Get<TResponse>(string url, HeadersCollection headers = null)
        {
            using (var client = GetClient())
            {
                HttpRequestMessage request = new HttpRequestMessage(HttpMethod.Get, url);
                if (headers != null)
                {
                    foreach (var header in headers)
                    {
                        request.Headers.Add(header.Key, header.Value);
                    }
                }
                request.Headers.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
                var responseMessage = client.SendAsync(request).Result;

                //task.Wait();
                //bool responseSuccessful = false;
                TResponse response = responseMessage.Content.ReadAsAsync<TResponse>().Result;
                return response;
            }
        }

        public async Task<TResponse> PostWithHeaders<TRequest, TResponse>(string url, TRequest request, List<KeyValuePair<string, string>> headers)
        {
            var response = Post<TRequest>(request, url, headers);
            JsonMediaTypeFormatter formatter = new JsonMediaTypeFormatter();
            return response.Content.ReadAsAsync<TResponse>().Result;
        }

        public HttpClient GetClient()
        {
            HttpClient client = new HttpClient();
            client.Timeout = TimeSpan.FromMinutes(10);
            client.BaseAddress = new Uri(SERVICE_URL);
            return client;
        }

        public TResponse Post<TResponse>(JToken obj, string url, HeadersCollection headers)
        {
            using (HttpRequestMessage request = new HttpRequestMessage(HttpMethod.Post, url))
            using (var client = GetClient())
            {
                var str = obj.ToString();
                request.Content = new StringContent(str, Encoding.UTF8, "application/json");
                if (headers != null)
                {
                    foreach (var header in headers)
                    {
                        request.Headers.Add(header.Key, header.Value);
                    }
                }
                request.Headers.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
                var response = client.SendAsync(request).Result;
                JsonMediaTypeFormatter formatter = new JsonMediaTypeFormatter();
                return response.Content.ReadAsAsync<TResponse>().Result;
            }
        }

        public TResponse Put<TResponse>(JToken obj, string url, HeadersCollection headers)
        {
            using (HttpRequestMessage request = new HttpRequestMessage(HttpMethod.Put, url))
            using (var client = GetClient())
            {
                var str = obj.ToString();
                request.Content = new StringContent(str, Encoding.UTF8, "application/json");
                if (headers != null)
                {
                    foreach (var header in headers)
                    {
                        request.Headers.Add(header.Key, header.Value);
                    }
                }
                request.Headers.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
                var response = client.SendAsync(request).Result;
                JsonMediaTypeFormatter formatter = new JsonMediaTypeFormatter();
                return response.Content.ReadAsAsync<TResponse>().Result;
            }
        }

        private HttpResponseMessage Post<T>(T obj, string url, List<KeyValuePair<string, string>> headers)
        {
            using (HttpRequestMessage request = new HttpRequestMessage(HttpMethod.Post, url))
            using (var client = GetClient())
            {
                JsonMediaTypeFormatter formatter = new JsonMediaTypeFormatter();
                var str = Serialize<T>(formatter, obj);
                request.Content = new StringContent(str, Encoding.UTF8, "application/json");
                if (headers != null)
                {
                    foreach (var header in headers)
                    {
                        request.Headers.Add(header.Key, header.Value);
                    }
                }
                request.Headers.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
                var response = client.SendAsync(request).Result;
                return response;
            }
        }

        public HttpResponseMessage Delete(string url, IDictionary<string, object> queryStringMap, HeadersCollection headers)
        {
            var queryString = QueryStringMap(queryStringMap);
            if (!string.IsNullOrEmpty(queryString))
            {
                url = url + "?" + queryString;
            }
            HttpRequestMessage request = new HttpRequestMessage(HttpMethod.Delete, url);
            if (headers != null)
            {
                foreach (var header in headers)
                {
                    request.Headers.Add(header.Key, header.Value);
                }
            }
            var response = GetClient().SendAsync(request).Result;
            return response;
        }

        string Serialize<T>(MediaTypeFormatter formatter, T value)
        {
            // Create a dummy HTTP Content.
            Stream stream = new MemoryStream();
            var content = new StreamContent(stream);
            // Serialize the object.
            formatter.WriteToStreamAsync(typeof(T), value, stream, content, null).Wait();
            // Read the serialized string.
            stream.Position = 0;
            return content.ReadAsStringAsync().Result;
        }

        private string QueryStringMap(IDictionary<string, object> queryStringMap)
        {
            var output = "";
            if (queryStringMap != null)
            {
                foreach (var entry in queryStringMap)
                {
                    var key = entry.Key;
                    var value = entry.Value;
                    if (value == null) continue;
                    if (!IsCollection(value))
                    {
                        output += (string.IsNullOrEmpty(output) ? "" : "&") + key + "=" + value;
                    }
                    else
                    {
                        foreach (var item in (IEnumerable)value)
                        {
                            output += string.IsNullOrEmpty(output) ? "" : "&" + key + "=" + item;
                        }
                    }
                }
            }
            return output;
        }

        public bool IsCollection(object instance)
        {
            if (instance == null) return false;
            var type = instance.GetType();
            if (type != typeof(string) && typeof(IEnumerable).IsAssignableFrom(type))
            {
                return true;
            }
            return false;
        }
    }

    public class HeadersCollection : IEnumerable<KeyValuePair<string, string>>, IEnumerator<KeyValuePair<string, string>>
    {
        IEnumerator<KeyValuePair<string, string>> _currentEnumerator;
        public HeadersCollection()
        {
            _collection = new List<KeyValuePair<string, string>>();
        }
        List<KeyValuePair<string, string>> _collection;

        public void Add(string key, string value)
        {
            _collection.Add(new KeyValuePair<string, string>(key, value));
        }

        public IEnumerator<KeyValuePair<string, string>> GetEnumerator()
        {
            _currentEnumerator = _collection.GetEnumerator();
            return _currentEnumerator;
        }

        public KeyValuePair<string, string> Current
        {
            get { return _currentEnumerator.Current; }
        }

        public void Dispose()
        {
            if (_currentEnumerator != null)
            {
                _currentEnumerator.Dispose();
            }
        }

        object IEnumerator.Current
        {
            get { return _currentEnumerator.Current; }
        }

        public bool MoveNext()
        {
            return _currentEnumerator.MoveNext();
        }

        public void Reset()
        {
            _currentEnumerator.Dispose();
            GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }
    }
}
